#include <inc/lib.h>

#define debug		0

/********************************
 * FILE DESCRIPTOR MANIPULATORS *
 *                              *
 ********************************/

// Return the file descriptor pointer corresponding to fdnum.
//
// Returns NULL on error.  Errors are:
//   - if 'fdnum' is out of range (< 0 or >= MAXFD)
//   - if 'require_open' is true and the file descriptor is not open
Fd *Fd::fd(int fdnum, bool require_open)
{
	// File descriptors start at address FDTABLE and each one occupies
	// one page of memory (so they can be shared independently).
	
	// LAB 5: Your code here.

	panic("Fd::fd not implemented");
	return NULL;
}

// Finds and returns the smallest unused file descriptor.
//
// Returns the file descriptor on success, NULL on error.  The error case
// corresponds to:
//	-E_MAX_FD: no more file descriptors
Fd *Fd::unused_fd()
{
	// Hint: Use Fd::fd() and Fd::is_open().
	
	// LAB 5: Your code here.

	panic("Fd::unused_fd not implemented");
	return NULL;
}

// Returns true iff this file descriptor is open.
bool Fd::is_open() const
{
	// A file descriptor is open iff something is mapped there.
	return (vpd[PDX(this)] & PTE_P) && (vpt[PGNUM(this)] & PTE_P);
}


static Dev error_dev('e', "error");
extern Dev console_dev;
extern Dev file_dev;
extern Dev pipe_dev;

// Return the device associated with this file descriptor.
Dev *Fd::device() const
{
	if (fd_type == 'c')
		return &console_dev;
	else if (fd_type == 'p')
		return &pipe_dev;
	else if (fd_type == 'f')
		return &file_dev;
	else {
		cprintf("[%08x] unknown device type %d\n", env->env_id, fd_type);
		return &error_dev;
	}
}


/******************
 * FILE FUNCTIONS *
 *                *
 ******************/

int
close(int fdnum)
{
	if (Fd *fd = Fd::fd(fdnum, true)) {
		int r = fd->device()->close(fd);
		// Make sure fd is unmapped.  Might be a no-op if
		// device()->close(fd) already unmapped it.
		(void) sys_page_unmap(0, fd);
		return r;
	} else
		return -E_INVAL;
}

void
close_all()
{
	for (int i = 0; i < MAXFD; i++)
		close(i);
}

// Make file descriptor 'newfdnum' a duplicate of file descriptor 'oldfdnum'.
// For instance, writing onto either file descriptor will affect the
// file and the file offset of the other.
// Closes any previously open file descriptor at 'newfdnum'.
// This is implemented using virtual memory tricks (of course!).
int
dup(int oldfdnum, int newfdnum)
{
	int r;
	pte_t pte;
	Fd *oldfd, *newfd;

	if (!(oldfd = Fd::fd(oldfdnum, true))
	    || !(newfd = Fd::fd(newfdnum, false)))
		return -E_INVAL;
	close(newfdnum);

	char *ova = oldfd->data();
	char *nva = newfd->data();

	if (vpd[PDX(ova)] & PTE_P) {
		for (int i = 0; i < PTSIZE; i += PGSIZE) {
			pte = vpt[PGNUM(ova + i)];
			if (pte & PTE_P) {
				// should be no error here -- pd is already allocated
				if ((r = sys_page_map(0, ova + i, 0, nva + i, pte & PTE_USER)) < 0)
					goto err;
			}
		}
	}
	if ((r = sys_page_map(0, oldfd, 0, newfd, vpt[PGNUM(oldfd)] & PTE_USER)) < 0)
		goto err;

	return newfdnum;

err:
	sys_page_unmap(0, newfd);
	for (int i = 0; i < PTSIZE; i += PGSIZE)
		sys_page_unmap(0, nva + i);
	return r;
}

ssize_t
read(int fdnum, void *buf, size_t n)
{
	Fd *fd;

	if (!(fd = Fd::fd(fdnum, true)))
		return -E_INVAL;
	if ((fd->fd_omode & O_ACCMODE) == O_WRONLY) {
		cprintf("[%08x] read %d -- bad mode\n", env->env_id, fdnum); 
		return -E_INVAL;
	}
	int r = fd->device()->read(fd, buf, n, fd->fd_offset);
	if (r >= 0)
		fd->fd_offset += r;
	return r;
}

ssize_t
readn(int fdnum, void *buf, size_t n)
{
	size_t tot, m;
	for (tot = 0; tot < n; tot += m) {
		m = read(fdnum, (char *) buf + tot, n - tot);
		if (m < 0)
			return m;
		if (m == 0)
			break;
	}
	return tot;
}

ssize_t
write(int fdnum, const void *buf, size_t n)
{
	Fd *fd;

	if (!(fd = Fd::fd(fdnum, true)))
		return -E_INVAL;
	if ((fd->fd_omode & O_ACCMODE) == O_RDONLY) {
		cprintf("[%08x] write %d -- bad mode\n", env->env_id, fdnum);
		return -E_INVAL;
	}
	if (debug)
		cprintf("write %d %p %d via dev %s\n",
			fdnum, buf, n, fd->device()->dev_name);
	int r = fd->device()->write(fd, buf, n, fd->fd_offset);
	if (r > 0)
		fd->fd_offset += r;
	return r;
}

int
seek(int fdnum, off_t offset)
{
	if (Fd *fd = Fd::fd(fdnum, true)) {
		fd->fd_offset = offset;
		return 0;
	} else
		return -E_INVAL;
}

int
ftruncate(int fdnum, off_t newsize)
{
	Fd *fd;
	if (!(fd = Fd::fd(fdnum, true)))
		return -E_INVAL;
	if ((fd->fd_omode & O_ACCMODE) == O_RDONLY) {
		cprintf("[%08x] ftruncate %d -- bad mode\n",
			env->env_id, fdnum); 
		return -E_INVAL;
	}
	return fd->device()->trunc(fd, newsize);
}

int
fstat(int fdnum, struct Stat *stat)
{
	Fd *fd;

	if (!(fd = Fd::fd(fdnum, true)))
		return -E_INVAL;
	stat->st_name[0] = 0;
	stat->st_size = 0;
	stat->st_isdir = 0;
	stat->st_dev = fd->device();
	return fd->device()->stat(fd, stat);
}

int
stat(const char *path, struct Stat *stat)
{
	int fd;

	if ((fd = open(path, O_RDONLY)) < 0)
		return fd;
	int r = fstat(fd, stat);
	close(fd);
	return r;
}



/**************************
 * BASIC DEVICE FUNCTIONS *
 *                        *
 **************************/

Dev::Dev(int dev_id_, const char *dev_name_)
	: dev_id(dev_id_), dev_name(dev_name_)
{
}

ssize_t Dev::read(struct Fd *fd, void *buf, size_t len, off_t offset)
{
	return -E_INVAL;
}

ssize_t Dev::write(struct Fd *fd, const void *buf, size_t len, off_t offset)
{
	return -E_INVAL;
}

int Dev::close(struct Fd *fd)
{
	return -E_INVAL;
}

int Dev::stat(struct Fd *fd, struct Stat *stat)
{
	return -E_INVAL;
}

int Dev::trunc(struct Fd *fd, off_t len)
{
	return -E_INVAL;
}

